const gql = require('graphql-tag')

const changeCase = require('change-case')
const inflection = require( 'inflection' )

const yves = require('yves')

const _ = require('lodash')
const globToRegExp = require('glob-to-regexp')

const directQueries = false
const pkg = require('../package.json')
const fs = require('fs')
const path = require('path')
const { indexOf } = require('lodash')

function moduleSetup(moduleOptions,moduleName) {
  const debug = yves.debugger(`${pkg.name}:${moduleName}`)

  let moduleSchema
  const rootName = moduleOptions.root

  if (moduleOptions.enabled && moduleOptions.schema) {
    debug('input schema %y', moduleOptions.schema)
    moduleSchema = require(moduleOptions.schema)
  }
  const INTROSPECT = 'introspect'
  const TYPE_INTROSPECT = `${rootName}__${changeCase.pascalCase(INTROSPECT)}`

  const resolvers = {
    Query: {
      [rootName.toLowerCase()]: (_source, args, context ) => {
        if (context.dataSources[moduleName]) {
          context.dataSources[moduleName].schema = moduleSchema
          return context.dataSources[moduleName].access(args.token, context)
        }
      },
    },
    [rootName]: {},
  }

  function prefix(name)
  {
    if (!name) return ''
    const l=name.toLowerCase().trim()
    if (l==='int' || l==='float' || l === 'string' || l==='boolean' || l === 'id' || l==='object' || l==='date') return name
    return `${rootName}_${name}`
  }

  let gqlSchema = `
extend type Query {
  "${rootName} accessor"
  ${rootName.toLowerCase()}${moduleOptions.authentication && moduleOptions.authentication.enabled && moduleOptions.authentication.type === 'token' ? '(token: String!)':''}: ${rootName}
}

type ${rootName} {
${partAccessors()}}

type ${TYPE_INTROSPECT} {

  "MongoDB collection"
  collection: String

  "MongoDB filter"
  filter: Object

  "MongoDB skip option"
  skip: Int

  "MongoDB limit option"
  limit: Int

  "MongoDB sort option"
  sort: Object

  "MongoDB options"
  options: Object

  "Result item count"
  count: Int

  "Total item count, not skipped nor limited"
  total: Int

  "MongoDB result"
  data: [ Object ]

  "MongoDB explanation (show vital index information)"
  explain: Object
}

`


  function makeParameterSpec(collection,requiredFields) {
    let result=''
    const schema = _.get(moduleSchema,`collections.${collection}.schema`)
    if (schema) {
      for (let field in schema) {
        if (schema[field].type == 'string') {
          result += ` "Filter field '${field}'" ${field}:String`
        } else if (schema[field].type == 'integer') {
          result += ` "Filter field '${field}'" ${field}:Int`
        } else if (schema[field].type == 'float') {
          result += ` "Filter field '${field}'" ${field}:Float`
        } else if (schema[field].type == 'boolean') {
          result += ` "Filter field '${field}'" ${field}:Boolean`
        }
        if (requiredFields && requiredFields.indexOf(field)>=0) {
          result += '!'
        }
      }
    }
    const textIndex = _.get(moduleSchema,`indexes.${collection}`)
    result += ' "Adds an extra MongoDB style filter" _filter: Object'
    if (textIndex) result += ' "Searches the text fields for matches" _search:String'
    result += ' "Limits the number of results, defaults to ${moduleOptions.limit}" _limit: Int'
    result += ' "Skips number of results, defaults to 0" _skip: Int, "Sort instruction MongoDB style" _sort: Object'
    result += ' "Extra MongoDB options" _options: Object'

    return result
  }

  function makeFilterWithArgs(args,collection) {
    function sensibleObject(filter) {
      return filter && (typeof filter == 'object') && (Object.keys(filter).length > 0)
    }
  
    function mergeFilters(...queries) {
      let result
      const trueQueries = queries.filter( query => sensibleObject(query) )
      if (trueQueries.length == 1) {
        result = trueQueries[0]
      } else if (trueQueries.length > 1) {
        result = { $and: trueQueries }
      } else {
        result = {}
      }
      return result
    }
  
    let $text
    if (args._search) {
      $text={$text: { $search: args._search, $language:'none' } }
    }
    const adapterArgs = ['_filter','_search','_limit','_skip','_sort','_options']
    const fieldsFilter={}
    const fieldArgs = Object.keys(args).filter( arg => adapterArgs.indexOf(arg)<0)
    fieldArgs.forEach(arg => {
      if (_.get(moduleSchema,`collections.${collection}.schema.${arg}.type`)=='string' && typeof args[arg] == 'string' && args[arg].indexOf('*')>=0) {
        fieldsFilter[arg]={ $regex: globToRegExp(args[arg]), $options: 'i' }
      } else {
        fieldsFilter[arg]=args[arg]
      }
     })
    return mergeFilters(args._filter,fieldsFilter,$text)
  }


  function makeOptionsWithArgs(args) {
    function cleanObject(obj) {
      if (typeof(obj) == 'object') Object.keys(obj).forEach(function(key) {if (obj[key] === undefined) delete obj[key]} )
      return obj || {}
    }

    function makeOptions(options,obj) {
      return Object.assign( {}, cleanObject(options), cleanObject(obj))
    }

    return makeOptions(args._options,{limit:args._limit,skip:args._skip,sort:args._sort})
  }

  function graphQLmap(schema, container, description, path, title, level = 0) {
    const typeEntries = []
    const unionEntries = []

    for (let entry in schema[container]) {
      const type = schema[container][entry].type
      const oSchema = schema[container][entry].schema

      if ((type == 'collection' || type == 'object') && oSchema && Object.keys(oSchema).length) {
        const mTitle = changeCase.pascalCase( (title ? `${title} ` : '') + inflection.singularize(entry) )
        graphQLmap(schema[container][entry],'schema', schema[container][entry].description, path ? `${path}.${entry}` : entry, mTitle, level + 1)
        if (type == 'object') typeEntries.push(`  ${entry.replace(/\W/g,'')}: ${prefix(mTitle)}`)
      } else if (type == 'array') {
        const memberType = schema[container][entry].member.type
        const memberSchema = schema[container][entry].member.schema
        let mTitle = changeCase.pascalCase(memberType == 'object' && (!memberSchema || Object.keys(memberSchema).length == 0) ? 'object' : memberType.replace('integer','int'))
        if (memberType == 'object' && memberSchema && Object.keys(memberSchema).length) {
          mTitle = changeCase.pascalCase( (title ? `${title} ` : '') + inflection.singularize(entry) )
          graphQLmap(schema[container][entry].member,'schema', null,path ? `${path}.${entry}` : entry, mTitle, level + 1)
        }
        const reference = schema[container][entry].member.reference
        if (reference) {
          if (!resolvers[prefix(title)]) resolvers[prefix(title)] = {}
          resolvers[prefix(title)][entry] = (source, args, context) => {
            if (context.dataSources[moduleName]) {
              if (directQueries) {
                const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,reference))
                return context.dataSources[moduleName].query(reference,filter,makeOptionsWithArgs(args),context)
              } else {
                return context.dataSources[moduleName].getMany(source[entry],reference,makeFilterWithArgs(args,reference),makeOptionsWithArgs(args),context)
              }
            }
          }
          typeEntries.push(`  ${entry.replace(/\W/g,'')}(${makeParameterSpec(reference)}): [ ${prefix(changeCase.pascalCase(inflection.singularize(reference)))} ]`)

          if (moduleOptions.introspect.enabled) {
            resolvers[prefix(title)][`${INTROSPECT}_${entry}`] = (source, args, context, info) => {
              const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,reference))
              if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(reference,filter, makeOptionsWithArgs(args), context, info)
            }
            const IAentry = `  ${INTROSPECT}_${entry}(${makeParameterSpec(reference)}): ${TYPE_INTROSPECT}`
            if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
          }
        } else {
          const references = schema[container][entry].member.references
          if (references) {

            const unionName = `${prefix(title)}_${entry.replace(/\W/g,'')}`
            unionEntries.push(`union ${unionName} = ${references.map( reference => prefix(changeCase.pascalCase(inflection.singularize(reference)))).join(' | ') }`)

            if (!resolvers[prefix(title)]) resolvers[prefix(title)] = {}

            resolvers[prefix(title)][entry] = (source, args, context, info) => {
              // See https://stackoverflow.com/a/46313828
              const requestedCollections = _.get(info,'fieldNodes[0].selectionSet.selections',[])
                .map(s => _.get(s,'typeCondition.name.value'))
                .filter( e => !!e )
                .map( type => inflection.pluralize(type.substr(moduleName.length+1).toLowerCase()) )  // reverse map GraphQL Type to mongo collection name
              if (context.dataSources[moduleName]) {
                if (Array.isArray(source[entry]) && source[entry].length) {
                  const subQueries = references
                    .filter( reference => !requestedCollections || requestedCollections.indexOf(reference)>=0)
                    .map( reference => {
                      if (directQueries) {
                        const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,reference))
                        return context.dataSources[moduleName].query(reference,filter,makeOptionsWithArgs(args),context)
                      } else {
                        return context.dataSources[moduleName].getMany(source[entry],reference,makeFilterWithArgs(args,reference),makeOptionsWithArgs(args),context)
                      }
                    })

                  return Promise.all(subQueries).then( resultArray => {
                    const result = [].concat.apply([], resultArray).filter( r => !!r )

                    const itemsByIds = _.keyBy( result, '_id')
                    return source[entry].map(id => (itemsByIds && itemsByIds[id]) || null).filter( r => !!r )
                  })
                } else {
                  return []
                }
              }
            }

            typeEntries.push(`  ${entry.replace(/\W/g,'')}(${makeParameterSpec(references)}): [ ${ unionName } ]`)

            if (moduleOptions.introspect.enabled) {
              resolvers[prefix(title)][`${INTROSPECT}_${entry.replace(/\W/g,'')}`] = (source, args, context, info) => {
                const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args))
                if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspects(references, filter, makeOptionsWithArgs(args), context, info)
              }
              const IAentry = `  ${INTROSPECT}_${entry.replace(/\W/g,'')}(${makeParameterSpec(references)}): [ ${TYPE_INTROSPECT} ]`
              if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
            }
  
            resolvers[unionName] = {
              __resolveType(entry) {
                return `${changeCase.pascalCase(moduleName)}_${changeCase.pascalCase(inflection.singularize(entry.__collection))}`
              }
            }

          } else {
            typeEntries.push(`  ${entry.replace(/\W/g,'')}: [ ${prefix(mTitle)} ]`)
          }
        }
      } else {
        const reference = (schema[container][entry].reference && moduleSchema.collections[schema[container][entry].reference]) ? schema[container][entry].reference : null
        if (reference) {
          const refType = changeCase.pascalCase(inflection.singularize(reference))

          if (!resolvers[prefix(title)]) resolvers[prefix(title)]={}

          resolvers[prefix(title)][entry.replace(/\W/g,'')] = (source, args, context) => {
            if (context.dataSources[moduleName]) return context.dataSources[moduleName].getSingle(source[entry], reference, makeFilterWithArgs(args), makeOptionsWithArgs(args),context)
          }
          typeEntries.push(`  ${entry.replace(/\W/g,'')}: ${prefix(refType)}`)

          if (moduleOptions.introspect.enabled) {
            resolvers[prefix(title)][`${INTROSPECT}_${entry.replace(/\W/g,'')}`] = (source, args, context, info) => {
              const filter = context.dataSources[moduleName].mergeFilters({_id:source[entry]},makeFilterWithArgs(args,reference))
              if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(reference, filter, makeOptionsWithArgs(args), context, info)
            }
            const IAentry = `  ${INTROSPECT}_${entry.replace(/\W/g,'')}(${makeParameterSpec(reference)}): ${TYPE_INTROSPECT}`
            if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
          }
        } else {
          const oSchema = schema[container][entry].schema
          typeEntries.push(`  ${entry.replace(/\W/g,'')}: ${prefix(changeCase.pascalCase(type == 'object' && (!oSchema || Object.keys(oSchema).length == 0) ? 'object' : type.replace('integer','int')))}`)
        }
      }
    }




    if (level == 1) {
      for (let refEntry in moduleSchema.collections) {
        for (let field in moduleSchema.collections[refEntry].schema) {
          const type = moduleSchema.collections[refEntry].schema[field].type
          const reference = moduleSchema.collections[refEntry].schema[field].reference || (type == 'array' && moduleSchema.collections[refEntry].schema[field].member && moduleSchema.collections[refEntry].schema[field].member.reference)
          if (reference == path && moduleSchema.collections[reference]) {
            const reverse = moduleSchema.collections[refEntry].schema[field].reverse || (type == 'array' && moduleSchema.collections[refEntry].schema[field].member && moduleSchema.collections[refEntry].schema[field].member.reverse)
            const mainType = changeCase.pascalCase(inflection.singularize(path))
            if (!resolvers[prefix(mainType)]) resolvers[prefix(mainType)]={}
            let virtField = reverse || inflection.pluralize(refEntry).replace(/\W/g,'')
            if (resolvers[prefix(mainType)][virtField]) virtField=field.toLowerCase()+'_'+virtField

            resolvers[prefix(mainType)][virtField] = (source, args, context) => {
              // if (/*1 || */directQueries) {
              const filter = context.dataSources[moduleName].mergeFilters({[field]:source._id},makeFilterWithArgs(args,refEntry))
              return context.dataSources[moduleName].query(refEntry,filter,makeOptionsWithArgs(args),context)
              // } else {
              //   const filter = context.dataSources[moduleName].mergeFilters({[field]:source._id},makeFilterWithArgs(args,refEntry))
              //   return context.dataSources[moduleName].getMany([], refEntry, filter, makeOptionsWithArgs(args), context)
              // }

            }
            const gentry = `  ${virtField}(${makeParameterSpec(refEntry)}): [ ${prefix(changeCase.pascalCase(inflection.singularize(refEntry)))} ]`
            if (typeEntries.indexOf(gentry)<0) typeEntries.push(gentry)
            if (moduleOptions.introspect.enabled) {
              resolvers[prefix(mainType)][`${INTROSPECT}_${virtField}`] = (source, args, context, info) => {
                const filter = context.dataSources[moduleName].mergeFilters({[field]:source._id},makeFilterWithArgs(args,refEntry))
                if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(refEntry,filter, makeOptionsWithArgs(args), context, info)
              }
              const IAentry = `  ${INTROSPECT}_${virtField}(${makeParameterSpec(refEntry)}): ${TYPE_INTROSPECT}`
              if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
            }
          }
        }
      }
    }


    if (typeEntries.length) {
      const pat = new RegExp(`^\\s*${INTROSPECT}_`, '')

      const unions = unionEntries.join('\n')
      if (unions) gqlSchema += unions+'\n\n'

      if (description) gqlSchema += `"${description}"\n`

      gqlSchema += `type ${prefix(title)} {\n${typeEntries.sort((a, b) => {
        const m_a = a.match(pat)
        const m_b = b.match(pat)
        if (m_a != m_b) {
          if (m_a) return 1
          if (m_b) return -1
        }
        return 0
      }).join('\n')}\n}\n\n`
    }
  }


  if (moduleSchema) graphQLmap(moduleSchema,'collections')

  function partAccessors() {
    let result=''
    let result2=''
    if (moduleSchema) {
      const collections = Object.keys(moduleSchema.collections)

      collections.forEach( collection => {
        result += `  ${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec(collection,['_id'])/*_id: String!*/}): ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))}\n`
        result += `  ${inflection.pluralize(collection).replace(/\W/g,'')}(${makeParameterSpec(collection)}): [ ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))} ]\n`

        if (moduleOptions.introspect.enabled) {
          result2 += `  ${INTROSPECT}_${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec(collection)}): ${TYPE_INTROSPECT}\n`
          result2 += `  ${INTROSPECT}_${inflection.pluralize(collection).replace(/\W/g,'')}(${makeParameterSpec(collection)}): ${TYPE_INTROSPECT}\n`
        }
      })
    }
    return result+result2
  }

  if (moduleOptions.writeGraphQLfile) {
    const filename=path.join(path.dirname(moduleOptions.schema),`${rootName.toLowerCase()}.graphql`)
    fs.writeFileSync(filename, gqlSchema)
    debug('output schema %y',filename)
  }

  const schema = moduleSchema ? gql([gqlSchema]) : null


  if (moduleSchema) {
    const collections = Object.keys(moduleSchema.collections)
    collections.forEach( collection => {
      resolvers[rootName][inflection.singularize(collection).replace(/\W/g,'')] = (_source, args, context) => {
        const filter=makeFilterWithArgs(args,collection)
        delete(filter._id)
        if (context.dataSources[moduleName]) return context.dataSources[moduleName].getSingle(args._id, collection, filter, makeOptionsWithArgs(args), context)
      }
      resolvers[rootName][inflection.pluralize(collection).replace(/\W/g,'')] = (_source, args, context) => {
        if (context.dataSources[moduleName]) return context.dataSources[moduleName].query(collection,makeFilterWithArgs(args,collection), makeOptionsWithArgs(args), context)
      }

      if (moduleOptions.introspect.enabled) {
        resolvers[rootName][`${INTROSPECT}_${inflection.singularize(collection).replace(/\W/g,'')}`] = (_source, args, context, info) => {
          if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(collection, makeFilterWithArgs(args,collection), makeOptionsWithArgs(args), context, info)
        }
        resolvers[rootName][`${INTROSPECT}_${inflection.pluralize(collection).replace(/\W/g,'')}`] = (_source, args, context, info) => {
          if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(collection, makeFilterWithArgs(args,collection), makeOptionsWithArgs(args), context, info)
        }
      }
    })
  }

  return moduleSchema ? {
    schema,
    resolvers,
  } : null
}

module.exports = {
  moduleSetup
}
